////////////////////////////////////////////////////////////////////////////////
// FastGraphicLib version 1.0
// ==========================
//
// A Fastest Win32.DIB bitmap graphic engine
//           .. for modern delphi languages.
//
//  This is a Delphi Library combined these library for compatibles for
// modern Delphi.
//
//    - TFast256
//    - TFastBMP
//    - TFastDraw
//    - TFastFX
//    - TFastIMG
//
// (C)2010 rageworx softwares , rageworx@gmail.com
//         programmed by Min Chul, KIM
//         2010.10.06
//
// * this code from FastLib of Gordon Alex Cowie
//   FastLib :
//             (C)1999 Gordon Alex Cowie <gfody@jps.net>
//             http://www.jps.net/gfody (not availed in this time)
//
// * Left to do
//
//     1) LoadFromStream(x:TStream);
//
//
////////////////////////////////////////////////////////////////////////////////
unit FastGraphicLib;

interface

uses
    Windows, SysUtils, Controls, Messages, Classes;

type
	// FASTRGB _________________________________________________________________
	TFColor		= record
		b,g,r	: byte;
	end;
    PFColor     = ^TFColor;

    TLine       = array[0..0] of TFColor;
    PLine       = ^TLine;

    TPLines     = array[0..0] of PLine;
    PPLines     = ^TPLines;

    TFastRGB = class
        public
            Gap,            // RowInc - Width
            RowInc,         // width in bytes
            Size,           // size of Bits
            Width,          // width in pixels
            Height      : Integer;
            Pixels      : PPLines;
            Bits        : Pointer;

        public
            procedure SetSize(fWidth,fHeight:Integer); virtual; abstract;
            procedure Draw(fdc,x,y:Integer); virtual; abstract;
            procedure Stretch(fdc,x,y,w,h:Integer); virtual; abstract;
            procedure DrawRect(fdc,x,y,w,h,sx,sy:Integer); virtual; abstract;
            procedure StretchRect(fdc,x,y,w,h,sx,sy,sw,sh:Integer); virtual; abstract;
            procedure TileDraw(fdc,x,y,w,h:Integer); virtual; abstract;
            procedure Resize(Dst:TFastRGB); virtual; abstract;
            procedure SmoothResize(Dst:TFastRGB); virtual; abstract;
            procedure CopyRect(Dst:TFastRGB;x,y,w,h,sx,sy:Integer); virtual; abstract;
            procedure Tile(Dst:TFastRGB); virtual; abstract;
    end;

	// FASTBMP _________________________________________________________________
    TFastBMP=class(TFastRGB)
        public
            hDC,
            Handle:     Integer;
            bmInfo:     TBitmapInfo;

        public
            constructor Create;
            destructor  Destroy; override;
            // initializers
            procedure   SetSize(fWidth,fHeight:Integer); override;
            procedure   LoadFromFile(FileName:string);
            procedure   SaveToFile(FileName:string);
            procedure   LoadFromhBmp(hBmp:Integer);
            procedure   LoadFromRes(hInst:Integer;sName:string);
            procedure   SetInterface(fWidth,fHeight:Integer;pBits:Pointer);
            procedure   Copy(FB:TFastBMP);
            // GDI drawing methods
            procedure   Draw(fdc,x,y:Integer); override;
            procedure   Stretch(fdc,x,y,w,h:Integer); override;
            procedure   DrawRect(fdc,x,y,w,h,sx,sy:Integer); override;
            procedure   StretchRect(fdc,x,y,w,h,sx,sy,sw,sh:Integer); override;
            procedure   TileDraw(fdc,x,y,w,h:Integer); override;
            // native conversions
            procedure   Resize(Dst:TFastRGB); override;
            procedure   SmoothResize(Dst:TFastRGB); override;
            procedure   CopyRect(Dst:TFastRGB;x,y,w,h,sx,sy:Integer); override;
            procedure   Tile(Dst:TFastRGB); override;
            function    CountColors:Longint;
    end;

	// FAST256 _________________________________________________________________
    TConversionMode   = (cmGray,cmCut,cmDither);

    TFColorEntry      = record
        b,
        g,
        r,
        a:Byte;
    end;

    PFColorEntry      = ^TFColorEntry;
    TFColorEntryArray = array[0..255]of TFColorEntry;
    PFColorEntryArray = ^TFColorEntryArray;

    TBitmapInfo256 = record
      bmiHeader: TBitmapInfoHeader;
      bmiColors: TFColorEntryArray;
    end;

    TBytes  = array[0..0]of Byte;
    PBytes  = ^TBytes;
    TPBytes = array[0..0]of PBytes;
    PPBytes = ^TPBytes;

    TFast256=class(TFastRGB)
        public
            hDC,
            Handle:     Integer;
            bmInfo:     TBitmapInfo256;
            Colors:     PFColorEntryArray;
            Pixels:     PPBytes;

        public
            constructor Create;
            destructor  Destroy; override;
            // initializers
            procedure   SetSize(fWidth,fHeight:Integer); override;
            procedure   LoadFromFile(FileName:string;cm:TConversionMode);
            procedure   SaveToFile(FileName:string);
            procedure   LoadFromhBmp(hBmp:Integer;cm:TConversionMode);
            procedure   LoadFromRes(hInst:Integer;sName:string;cm:TConversionMode);
            procedure   LoadFromRGBExact(FB:TFastRGB);
            procedure   MakeGrayscale(FB:TFastRGB);
            procedure   LoadFromRGB(FB:TFastRGB;cm:TConversionMode);
            procedure   SetInterface(fWidth,fHeight:Integer;pBits:Pointer);
            procedure   Copy(FB:TFast256);
            // color table tools
            procedure   RGB(ra,ga,ba:Integer);
            procedure   Contrast(Amount:Integer);
            procedure   Saturation(Amount:Integer);
            procedure   Lightness(Amount:Integer);
            procedure   FillColors(c1,c255:TFColor);
            procedure   ShiftColors(Amount:Integer);
            function    MakePalette: HPALETTE;
            // GDI drawing methods
            procedure   Draw(fdc,x,y:Integer); override;
            procedure   Stretch(fdc,x,y,w,h:Integer); override;
            procedure   DrawRect(fdc,x,y,w,h,sx,sy:Integer); override;
            procedure   StretchRect(fdc,x,y,w,h,sx,sy,sw,sh:Integer); override;
            procedure   TileDraw(fdc,x,y,w,h:Integer); override;
            // native conversions
            procedure   CopyRect(Dst:TFastRGB;x,y,w,h,sx,sy:Integer); override;
            procedure   Tile(Dst:TFastRGB); override;
            procedure   Resize(Dst:TFastRGB); override;
            procedure   SmoothResize(Dst:TFastRGB); override;
    end;

    // FastGraphicLib __________________________________________________________
    TFastGraphic = class
        public
        	// FASTRGB _________________________________________________________
            function FRGB(r,g,b:Byte):TFColor;
            function IntToColor(i:Integer):TFColor;
            function IntToByte(i:Integer):Byte;
            function TrimInt(i,Min,Max:Integer):Integer;

        public
            // FAST256 _________________________________________________________
            function Check256(FB:TFastRGB):Boolean;
            function FRGBA(r,g,b,a:Byte):TFColorEntry;
            function ColorToEntry(tf:TFColor):TFColorEntry;
            function EntryToColor(tf:TFColorEntry):TFColor;

        public
            // FASTDRAW ________________________________________________________
            procedure Rectangle(Bmp:TFastRGB;x1,y1,x2,y2:Integer;clr:TFColor);
            procedure FillRect(Bmp:TFastRGB;x1,y1,x2,y2:Integer;clr:TFColor);
            procedure Line(Bmp:TFastRGB;x1,y1,x2,y2:Integer;clr:TFColor);
            procedure SmoothLine(Bmp:TFastRGB;x1,y1,x2,y2:Integer;clr:TFColor);
            procedure PolyLine(Bmp:TFastRGB;pnts:array of TPoint;clr:TFColor);
            procedure SmoothPolyLine(Bmp:TFastRGB;pnts:array of TPoint;clr:TFColor);
            procedure Ellipse(Bmp:TFastRGB;cx,cy,Rx,Ry:Integer;clr:TFColor);

        public
            // FASTIMG _________________________________________________________
            function  IsPalette:Boolean;
    end;

    // FastIMG _________________________________________________________________
    TBMPFilename = type string;
    
    TDrawStyle=(dsDraw,dsStretch,dsResize,dsTile,dsCenter);
    TDataType=(dtFastBMP,dtFast256);

    TFastIMG=class(TCustomControl)
        private
            fBmp        : TFastRGB;
            fDrawStyle  : TDrawStyle;
            fSmthSize,
            fAutoSize   : Boolean;
            fDataType   : TDataType;
            fCMode      : TConversionMode;
            fFileName   : TBMPFilename;

            procedure   SetBitmap(x:TFastRGB);
            procedure   SetSmthSize(x:Boolean);
            procedure   SetAutoSize(x:Boolean);
            procedure   SetDrawStyle(x:TDrawStyle);
            procedure   SetDataType(x:TDataType);
            procedure   SetConversionMode(x:TConversionMode);
            procedure   SetFileName(const x:TBMPFilename);
            procedure   Erase(var Msg:TWMEraseBkgnd); message WM_ERASEBKGND;
            procedure   Paint(var Msg:TWMPaint); message WM_PAINT;

        public
            constructor Create(AOwner:TComponent); override;
            destructor  Destroy; override;
            procedure   LoadFromFile(x:string);
            procedure   LoadFromResource(hInst:Integer;sName:string);
            procedure   LoadFromStream(x:TStream);
            property    Bmp:TFastRGB read fBmp write SetBitmap;

        published
            property    AutoSize:Boolean read fAutoSize write SetAutoSize;
            property    BilinearResize:Boolean read fSmthSize write SetSmthSize;
            property    DrawStyle:TDrawStyle read fDrawStyle write SetDrawStyle;
            property    DataType:TDataType read fDataType write SetDataType;
            property    Convert_256:TConversionMode read fCMode write SetConversionMode;
            property    FileName:TBMPFilename read fFileName write SetFileName;
            property    Align;
            property    Color;
            property    DragCursor;
            property    DragMode;
            property    Enabled;
            property    ParentColor;
            property    ParentFont;
            property    ParentShowHint;
            property    PopupMenu;
            property    ShowHint;
            property    Visible;
            property    OnClick;
            property    OnDblClick;
            property    OnDragDrop;
            property    OnDragOver;
            property    OnEndDrag;
            property    OnMouseDown;
            property    OnMouseMove;
            property    OnMouseUp;
            property    OnStartDrag;
    end;

const
	// FASTRGB _________________________________________________________________
    tfBlack   : TFColor=(b:0;g:0;r:0);
    tfMaroon  : TFColor=(b:0;g:0;r:128);
    tfGreen   : TFColor=(b:0;g:128;r:0);
    tfOlive   : TFColor=(b:0;g:128;r:128);
    tfNavy    : TFColor=(b:128;g:0;r:0);
    tfPurple  : TFColor=(b:128;g:0;r:128);
    tfTeal    : TFColor=(b:128;g:128;r:0);
    tfGray    : TFColor=(b:128;g:128;r:128);
    tfSilver  : TFColor=(b:192;g:192;r:192);
    tfRed     : TFColor=(b:0;g:0;r:255);
    tfLime    : TFColor=(b:0;g:255;r:0);
    tfYellow  : TFColor=(b:0;g:255;r:255);
    tfBlue    : TFColor=(b:255;g:0;r:0);
    tfFuchsia : TFColor=(b:255;g:0;r:255);
    tfAqua    : TFColor=(b:255;g:255;r:0);
    tfLtGray  : TFColor=(b:192;g:192;r:192);
    tfDkGray  : TFColor=(b:128;g:128;r:128);
    tfWhite   : TFColor=(b:255;g:255;r:255);

implementation

const
    hSection = 0;

var
    gTool   : TFastGraphic;

////////////////////////////////////////////////////////////////////////////////
// TFastGraphic_________________________________________________________________
function TFastGraphic.FRGB(r: Byte; g: Byte; b: Byte):TFColor;
begin
    Result.b:=b;
    Result.g:=g;
    Result.r:=r;
end;

function TFastGraphic.IntToColor(i: Integer):TFColor;
begin
    Result.b:=i shr 16;
    Result.g:=i shr 8;
    Result.r:=i;
end;

function TFastGraphic.IntToByte(i:Integer):Byte;
begin
    if (i>255) then
        Result:=255
    else
    if (i<0) then
        Result:=0
    else
        Result:=i;
end;

function TFastGraphic.TrimInt(i,Min,Max:Integer):Integer;
begin
    if (i>Max) then
        Result:=Max
    else
    if (i<Min) then
        Result:=Min
    else
        Result:=i;
end;

// checks if a TFastRGB has < 256 colors
function TFastGraphic.Check256(FB:TFastRGB):Boolean;
var
    xr,xg,xb,
    r,g,b,c     : Byte;
    x,y         : Integer;
    Tmp         : PFColor;
    Count       : Longint;
    Buff        : PByte;
    Space       : array[0..31,0..31,0..31]of Byte;
begin
    FillChar(Space,SizeOf(Space),0);
    Tmp:=FB.Bits;
    Count:=0;

    for y:=0 to FB.Height-1 do
    begin
        for x:=0 to FB.Width-1 do
        begin
            r:=Tmp.r shr 3;
            g:=Tmp.g shr 3;
            b:=Tmp.b shr 3;
            Buff:=@Space[r,g,b];
            xr:=1 shl(Tmp.r-(r shl 3));
            xg:=1 shl(Tmp.g-(g shl 3));
            xb:=1 shl(Tmp.b-(b shl 3));
            c:=Buff^ or xr or xg or xb;

            if (c>Buff^) then
            begin
                Inc(Count);
                if (Count=257) then
                begin
                    Result:=False;
                    Exit;
                end;
                Buff^:=c;
            end;
            Inc(Tmp);
        end;
        Tmp:=Ptr(Integer(Tmp)+FB.Gap);
    end;
    Result:=True;
end;

function TFastGraphic.FRGBA(r,g,b,a:Byte):TFColorEntry;
begin
    Result.b:=b;
    Result.g:=g;
    Result.r:=r;
    Result.a:=a;
end;

function TFastGraphic.EntryToColor(tf:TFColorEntry):TFColor;
begin
    Result.b:=tf.b;
    Result.g:=tf.g;
    Result.r:=tf.r;
end;

function TFastGraphic.ColorToEntry(tf:TFColor):TFColorEntry;
begin
    Result.b:=tf.b;
    Result.g:=tf.g;
    Result.r:=tf.r;
end;

procedure TFastGraphic.Rectangle(Bmp:TFastRGB;x1,y1,x2,y2:Integer;clr:TFColor);
var
    Tmp     : PFColor;
    i       : Integer;
begin
    if x1>x1 then begin i:=x1; x1:=x2; x2:=i; end;
    if y1>y2 then begin i:=y1; y1:=y2; y2:=i; end;

    if Bmp is TFast256 then
    begin
        for i:=x1 to x2 do
        begin
            TFast256(Bmp).Pixels[y1,i]:=0;
            TFast256(Bmp).Pixels[y2,i]:=0;
        end;

        for i:=y1 to y2 do
        begin
            TFast256(Bmp).Pixels[i,x1]:=0;
            TFast256(Bmp).Pixels[i,x2]:=0;
        end;
    end
    else
    begin
        for i:=x1 to x2 do
        begin
            Bmp.Pixels[y1,i]:=clr;
            Bmp.Pixels[y2,i]:=clr;
        end;
        
        for i:=y1 to y2 do
        begin
            Bmp.Pixels[i,x1]:=clr;
            Bmp.Pixels[i,x2]:=clr;
        end;
    end;
end;

procedure TFastGraphic.FillRect(Bmp:TFastRGB;x1,y1,x2,y2:Integer;clr:TFColor);
var
    i,x,y: Integer;
begin
    x1:=TrimInt(x1,0,Bmp.Width-1);
    x2:=TrimInt(x2,0,Bmp.Width-1);
    y1:=TrimInt(y1,0,Bmp.Height-1);
    y2:=TrimInt(y2,0,Bmp.Height-1);

    if (x1>x1) then
    begin
        i:=x1;
        x1:=x2;
        x2:=i;
    end;

    if (y1>y2) then
    begin
        i:=y1;
        y1:=y2;
        y2:=i;
    end;

    if Bmp is TFast256 then
        for y:=y1 to y2 do
            for x:=x1 to x2 do
                TFast256(Bmp).Pixels[y,x]:=0
    else
        for y:=y1 to y2 do
            for x:=x1 to x2 do
                Bmp.Pixels[y,x]:=clr;
end;

//bressenham line algorithm
procedure TFastGraphic.Line(Bmp:TFastRGB;x1,y1,x2,y2:Integer;clr:TFColor);
var
    d,
    ax,
    ay,
    sx,
    sy,
    dx,
    dy  : Integer;
begin
    if not((x1<Bmp.Width)and(x2<Bmp.Width)and(x1>-1)and(x2>-1)and
         (y1<Bmp.Height)and(y2<Bmp.Height)and(y1>-1)and(y2>-1))then
        Exit;

    dx:=x2-x1;
    ax:=Abs(dx)shl 1;

    if dx<0 then
        sx:=-1
    else
        sx:=1;

    dy:=y2-y1;
    ay:=Abs(dy)shl 1;

    if dy<0 then
        sy:=-1
    else
        sy:=1;

    Bmp.Pixels[y1,x1]:=clr;

    if ax>ay then
    begin
        d:=ay-(ax shr 1);

        while (x1<>x2) do
        begin
            if d>-1 then
            begin
                Inc(y1,sy);
                Dec(d,ax);
            end;

            Inc(x1,sx);
            Inc(d,ay);
            Bmp.Pixels[y1,x1]:=clr;
        end;
    end
    else
    begin
        d:=ax-(ay shr 1);

        while (y1<>y2) do
        begin
            if d>=0 then
            begin
                Inc(x1,sx);
                Dec(d,ay);
            end;
            
            Inc(y1,sy);
            Inc(d,ax);
            Bmp.Pixels[y1,x1]:=clr;
        end;
    end;
end;

//modified bressenham's to alphablend the error (antialiased)
procedure TFastGraphic.SmoothLine(Bmp:TFastRGB;x1,y1,x2,y2:Integer;clr:TFColor);
var
    ea,
    ec  : Word;
    ci  : Byte;
    dx,
    dy,
    d,s : Integer;
    Tmp : PFColor;
begin
    if(y1=y2)or(x1=x2)then
    begin
        Line(Bmp,x1,y1,x2,y2,clr);
        Exit;
    end;

    if y1>y2 then
    begin
        d:=y1; y1:=y2; y2:=d;
        d:=x1; x1:=x2; x2:=d;
    end;

    dx:=x2-x1;
    dy:=y2-y1;

    if dx>-1 then s:=1 else
    begin
        s:=-1;
        dx:=-dx;
    end;

    ec:=0;

    Bmp.Pixels[y1,x1]:=clr;

    if dy>dx then
    begin
        ea:=(dx shl 16)div dy;

        while dy>1 do
        begin
            Dec(dy);
            d:=ec;
            Inc(ec,ea);

            if ec<=d then
                Inc(x1,s);

            Inc(y1);
            ci:=ec shr 8;

            Tmp:=@Bmp.Pixels[y1,x1];
            Tmp.b:=((clr.b*(255 xor ci))+(Tmp.b*ci))shr 8;
            Tmp.g:=((clr.g*(255 xor ci))+(Tmp.g*ci))shr 8;
            Tmp.r:=((clr.r*(255 xor ci))+(Tmp.r*ci))shr 8;
            
            Tmp:=@Bmp.Pixels[y1,x1+s];
            Tmp.b:=((clr.b*ci)+(Tmp.b*(255 xor ci)))shr 8;
            Tmp.g:=((clr.g*ci)+(Tmp.g*(255 xor ci)))shr 8;
            Tmp.r:=((clr.r*ci)+(Tmp.r*(255 xor ci)))shr 8;
        end;
    end else
    begin
        ea:=(dy shl 16)div dx;
        while dx>1 do
        begin
            Dec(dx);
            d:=ec;
            Inc(ec,ea);
            
            if ec<=d then
                Inc(y1);

            Inc(x1,s);
            ci:=ec shr 8;

            Tmp:=@Bmp.Pixels[y1,x1];
            Tmp.b:=((clr.b*(255 xor ci))+(Tmp.b*ci))shr 8;
            Tmp.g:=((clr.g*(255 xor ci))+(Tmp.g*ci))shr 8;
            Tmp.r:=((clr.r*(255 xor ci))+(Tmp.r*ci))shr 8;

            Tmp:=@Bmp.Pixels[y1+1,x1];
            Tmp.b:=((clr.b*ci)+(Tmp.b*(255 xor ci)))shr 8;
            Tmp.g:=((clr.g*ci)+(Tmp.g*(255 xor ci)))shr 8;
            Tmp.r:=((clr.r*ci)+(Tmp.r*(255 xor ci)))shr 8;
        end;
    end;
    Bmp.Pixels[y2,x2]:=clr;
end;

procedure TFastGraphic.PolyLine(Bmp:TFastRGB;pnts:array of TPoint;clr:TFColor);
var
    n,i : Integer;
begin
    n:=High(pnts)+1;
    
    for i:=0 to n-1 do
        Line(Bmp,pnts[i].x,pnts[i].y,pnts[(i+1) mod n].x,pnts[(i+1) mod n].y,clr);
end;

procedure TFastGraphic.SmoothPolyLine(Bmp:TFastRGB;pnts:array of TPoint;clr:TFColor);
var
    n,i : Integer;
begin
  n:=High(pnts)+1;
  
  for i:=0 to n-1 do
      SmoothLine(Bmp,pnts[i].x,pnts[i].y,pnts[(i+1) mod n].x,pnts[(i+1) mod n].y,clr);
end;

procedure TFastGraphic.Ellipse(Bmp:TFastRGB;cx,cy,Rx,Ry:Integer;clr:TFColor);
var
    Rx2,
    Ry2,
    twoRx2,
    twoRy2,
    p,x,y,
    px,py   : Integer;
begin
    Rx2:=Rx*Rx;    Ry2:=Ry*Ry;
    twoRx2:=2*Rx2; twoRy2:=2*Ry2;
    x:=0;          y:=Ry;
    px:=0;         py:=twoRx2*y;

    Bmp.Pixels[cy+y,cx+x]:=clr;
    Bmp.Pixels[cy+y,cx-x]:=clr;
    Bmp.Pixels[cy-y,cx+x]:=clr;
    Bmp.Pixels[cy-y,cx-x]:=clr;

    p:=Ry2-(Rx2*Ry)+(Rx2 div 4);
    while px<py do
    begin
        Inc(x);
        Inc(px,twoRy2);

        if p<0 then Inc(p,Ry2+px)else
        begin
            Dec(y);
            Dec(py,twoRx2);
            Inc(p,Ry2+px-py);
        end;

        Bmp.Pixels[cy+y,cx+x]:=clr;
        Bmp.Pixels[cy+y,cx-x]:=clr;
        Bmp.Pixels[cy-y,cx+x]:=clr;
        Bmp.Pixels[cy-y,cx-x]:=clr;
    end;

    p:=Round(Ry2*(x+0.5)*(x+0.5)+Rx2*(y-1)*(y-1)-Rx2*Ry2);
    
    while y>0 do
    begin
        Dec(y);
        Dec(py,twoRx2);
        
        if p>0 then Inc(p,Rx2-py)else
        begin
            Inc(x);
            Inc(px,twoRy2);
            Inc(p,Rx2-py+px);
        end;

        Bmp.Pixels[cy+y,cx+x]:=clr;
        Bmp.Pixels[cy+y,cx-x]:=clr;
        Bmp.Pixels[cy-y,cx+x]:=clr;
        Bmp.Pixels[cy-y,cx-x]:=clr;
    end;
end;

function TFastGraphic.IsPalette:Boolean;
var
    sdc : HDC;
begin
    sdc:=GetDC(0);
    Result:=GetDeviceCaps(sdc,BITSPIXEL) < 9;
    ReleaseDC(0,sdc);
end;


////////////////////////////////////////////////////////////////////////////////
// FASTBMP _____________________________________________________________________
constructor TFastBMP.Create;
begin
    hDC:=CreateCompatibleDC(0);
    Width:=0;
    Height:=0;
    Bits:=nil;
    Pixels:=nil;

    bmInfo.bmiHeader.biSize:=SizeOf(TBitmapInfoHeader);
    bmInfo.bmiHeader.biPlanes:=1;
    bmInfo.bmiHeader.biBitCount:=24;
    bmInfo.bmiHeader.biCompression:=BI_RGB;
end;

destructor TFastBMP.Destroy;
begin
    DeleteDC(hDC);
    DeleteObject(Handle);
    FreeMem(Pixels);
    inherited Destroy;
end;

procedure TFastBMP.SetSize(fWidth,fHeight:Integer);
var
    x   : Longint;
    i   : Integer;
begin
    if(fWidth=Width)and(fHeight=Height)then
        Exit;

    Width:=fWidth;
    Height:=fHeight;
    bmInfo.bmiHeader.biWidth:=Width;
    bmInfo.bmiHeader.biHeight:=Height;
    Handle:=CreateDIBSection(0,bmInfo,DIB_RGB_COLORS,Bits,hSection,0);
    DeleteObject(SelectObject(hDC,Handle));
    RowInc:=((Width*24+31)shr 5)shl 2;
    Gap:=Width mod 4;
    Size:=RowInc*Height;
    GetMem(Pixels,Height*4);
    x:=Longint(Bits);

    for i:=0 to Height-1 do
    begin
        Pixels[i]:=Pointer(x);
        Inc(x,RowInc);
    end;
end;

procedure TFastBMP.LoadFromFile(FileName:string);
type
    TByteArray = array[0..0]of Byte;
    PByteArray =^TByteArray;
    TBMInfo256 = record
        bmiHeader: TBitmapInfoHeader;
        bmiColors: array[0..255]of TRGBQuad;
    end;
var
    hFile,
    bSize   : Integer;
    iRead   : Cardinal;
    fData   : TofStruct;
    fHead   : TBitmapFileHeader;
    fInfo   : TBMInfo256;
    fBits   : PByteArray;
begin
    hFile:=OpenFile(PChar(FileName),fData,OF_READ);
    ReadFile(hFile,fHead,SizeOf(fHead),iRead,nil);
    ReadFile(hFile,fInfo,SizeOf(fInfo),iRead,nil);
    SetFilePointer(hFile,fHead.bfOffBits,nil,FILE_BEGIN);
    SetSize(fInfo.bmiHeader.biWidth,Abs(fInfo.bmiHeader.biHeight));
    bSize:=(((fInfo.bmiHeader.biWidth*fInfo.bmiHeader.biBitCount+31)shr 5)shl 2)*fInfo.bmiHeader.biHeight;
    GetMem(fBits,bSize);
    ReadFile(hFile,fBits^,bSize,iRead,nil);
    StretchDIBits(hDC,0,0,Width,Height,0,0,Width,Height,fBits,PBitmapInfo(@fInfo)^,DIB_RGB_COLORS,SRCCOPY);
    FreeMem(fBits);
    CloseHandle(hFile);
end;

procedure TFastBMP.SaveToFile(FileName:string);
type
    TByteArray = array[0..0]of Byte;
    PByteArray =^TByteArray;
var
    hFile   : Integer;
    iWrote  : Cardinal;
    fData   : TofStruct;
    fHead   : TBitmapFileHeader;
begin
    // saves upside down.. just Flop from FastFX if you want
    hFile:=OpenFile(PChar(FileName),fData,OF_CREATE or OF_WRITE);
    fHead.bfType:=19778; // "BM"
    fHead.bfSize:=SizeOf(fHead)+SizeOf(bmInfo)+Size;
    fHead.bfOffBits:=SizeOf(fHead)+SizeOf(bmInfo);
    WriteFile(hFile,fHead,SizeOf(fHead),iWrote,nil);
    bmInfo.bmiHeader.biHeight:=Abs(bmInfo.bmiHeader.biHeight);
    WriteFile(hFile,bmInfo,SizeOf(bmInfo),iWrote,nil);
    bmInfo.bmiHeader.biHeight:=-bmInfo.bmiHeader.biHeight;
    WriteFile(hFile,PByteArray(Bits)^,Size,iWrote,nil);
    CloseHandle(hFile);
end;

procedure TFastBMP.LoadFromhBmp(hBmp:Integer);
var
    Bmp     : TBitmap;
    memDC   : Integer;
begin
    GetObject(hBmp,SizeOf(Bmp),@Bmp);
    SetSize(Bmp.bmWidth,Bmp.bmHeight);
    memDC:=CreateCompatibleDC(0);
    SelectObject(memDC,hBmp);
    GetDIBits(memDC,hBmp,0,Height,Bits,bmInfo,DIB_RGB_COLORS);
    DeleteDC(memDC);
end;

procedure TFastBMP.LoadFromRes(hInst:Integer;sName:string);
var
    hBmp    : Integer;
begin
    hBmp:=LoadImage(hInst,PChar(sName),IMAGE_BITMAP,0,0,0);
    LoadFromhBmp(hBmp);
    DeleteObject(hBmp);
end;

procedure TFastBMP.SetInterface(fWidth,fHeight:Integer;pBits:Pointer);
var
    x   : Longint;
    i   : Integer;
begin
    Width:=fWidth;
    Height:=Abs(fHeight);
    bmInfo.bmiHeader.biWidth:=fWidth;
    bmInfo.bmiHeader.biHeight:=fHeight;
    RowInc:=((Width*24+31)shr 5)shl 2;
    Gap:=Width mod 4;
    Size:=RowInc*Height;
    ReallocMem(Pixels,Height*4);
    Bits:=pBits;
    x:=Longint(Bits);
    
    for i:=0 to fHeight-1 do
    begin
        Pixels[i]:=Pointer(x);
        Inc(x,RowInc);
    end;
end;

procedure TFastBMP.Copy(FB:TFastBMP);
begin
    SetSize(FB.Width,FB.Height);
    CopyMemory(Bits,FB.Bits,Size);
end;

procedure TFastBMP.Draw(fdc,x,y:Integer);
begin
    StretchDIBits(fdc,x,y,Width,Height,0,0,Width,Height,Bits,bmInfo,DIB_RGB_COLORS,SRCCOPY);
end;

procedure TFastBMP.Stretch(fdc,x,y,w,h:Integer);
begin
    SetStretchBltMode(fdc,STRETCH_DELETESCANS);
    StretchDIBits(fdc,x,y,w,h,0,0,Width,Height,Bits,bmInfo,DIB_RGB_COLORS,SRCCOPY);
end;

procedure TFastBMP.DrawRect(fdc,x,y,w,h,sx,sy:Integer);
begin
    StretchDIBits(fdc,x,y,w,h,sx,sy,w,h,Bits,bmInfo,DIB_RGB_COLORS,SRCCOPY);
end;

procedure TFastBMP.StretchRect(fdc,x,y,w,h,sx,sy,sw,sh:Integer);
begin
    SetStretchBltMode(fdc,STRETCH_DELETESCANS);
    StretchDIBits(fdc,x,y,w,h,sx,sy,sw,sh,Bits,bmInfo,DIB_RGB_COLORS,SRCCOPY);
end;

procedure TFastBMP.TileDraw(fdc,x,y,w,h:Integer);
var
    wd,hd,
    hBmp,
    memDC: Integer;
begin
    if(Width=0)or(Height=0)then Exit;
    memDC:=CreateCompatibleDC(fdc);
    hBmp:=CreateCompatibleBitmap(fdc,w,h);
    SelectObject(memDC,hBmp);
    Draw(memDC,0,0);
    wd:=Width;
    hd:=Height;

    while wd<w do
    begin
        BitBlt(memDC,wd,0,wd*2,h,memDC,0,0,SRCCOPY);
        Inc(wd,wd);
    end;

    while hd<h do
    begin
        BitBlt(memDC,0,hd,w,hd*2,memDC,0,0,SRCCOPY);
        Inc(hd,hd);
    end;

    BitBlt(fdc,x,y,w,h,memDC,0,0,SRCCOPY);
    DeleteDC(memDC);
    DeleteObject(hBmp);
end;

procedure TFastBMP.Resize(Dst:TFastRGB);
var
    xCount,
    yCount,
    x,y,xP,yP,
    xD,yD,
    yiScale,
    xiScale     : Integer;
    xScale,
    yScale      : Single;
    Read,
    Line        : PLine;
    Tmp         : TFColor;
    pc          : PFColor;
begin
    if(Dst.Width=0)or(Dst.Height=0)then
        Exit;

    if(Dst.Width=Width)and(Dst.Height=Height)then
    begin
        CopyMemory(Dst.Bits,Bits,Size);
        Exit;
    end;

    xScale:=Dst.Width/Width;
    yScale:=Dst.Height/Height;
    if(xScale<1)or(yScale<1)then
    begin  // shrinking
        xiScale:=(Width shl 16) div Dst.Width;
        yiScale:=(Height shl 16) div Dst.Height;
        yP:=0;

        for y:=0 to Dst.Height-1 do
        begin
            xP:=0;
            read:=Pixels[yP shr 16];
            pc:=@Dst.Pixels[y,0];

            for x:=0 to Dst.Width-1 do
            begin
                pc^:=Read[xP shr 16];
                Inc(pc);
                Inc(xP,xiScale);
            end;

            Inc(yP,yiScale);
        end;
    end
    else   // zooming
    begin
        yiScale:=Round(yScale+0.5);
        xiScale:=Round(xScale+0.5);
        GetMem(Line,Dst.Width*3);

        for y:=0 to Height-1 do
        begin
            yP:=Trunc(yScale*y);
            Read:=Pixels[y];
            for x:=0 to Width-1 do
            begin
                xP:=Trunc(xScale*x);
                Tmp:=Read[x];
                for xCount:=0 to xiScale-1 do
                begin
                    xD:=xCount+xP;

                    if xD>=Dst.Width then
                        Break;

                    Line[xD]:=Tmp;
                end;
            end;

            for yCount:=0 to yiScale-1 do
            begin
                yD:=yCount+yP;
                if yD>=Dst.Height then Break;
                CopyMemory(Dst.Pixels[yD],Line,Dst.Width*3);
            end;
        end;

        FreeMem(Line);
    end;
end;

// huge thanks to Vit Kovalcik for this awesome function!
// performs a fast bilinear interpolation <vkovalcik@iname.com>
procedure TFastBMP.SmoothResize(Dst:TFastRGB);
var
    x,y,xP,yP,
    yP2,xP2     : Integer;
    Read,
    Read2       : PLine;
    t,z,z2,iz2  : Integer;
    pc          : PFColor;
    w1,w2,w3,w4 : Integer;
    Col1,Col2   : PFColor;
begin
    if (Dst.Width<1)or(Dst.Height<1) then
        Exit;

    if (Width=1) then
    begin
        Resize(Dst);
        Exit;
    end;

    if (Dst.Width=Width)and(Dst.Height=Height) then
    begin
        CopyMemory(Dst.Bits,Bits,Size);
        Exit;
    end;

    xP2:=((Width-1)shl 15)div Dst.Width;
    yP2:=((Height-1)shl 15)div Dst.Height;
    yP:=0;

    for y:=0 to Dst.Height-1 do
    begin
        xP:=0;
        Read:=Pixels[yP shr 15];

        if yP shr 16<Height-1 then
            Read2:=Pixels[yP shr 15+1]
        else
            Read2:=Pixels[yP shr 15];

        pc:=@Dst.Pixels[y,0];
        z2:=yP and $7FFF;
        iz2:=$8000-z2;

        for x:=0 to Dst.Width-1 do
        begin
            t:=xP shr 15;
            Col1:=@Read[t];
            Col2:=@Read2[t];
            z:=xP and $7FFF;
            w2:=(z*iz2)shr 15;
            w1:=iz2-w2;
            w4:=(z*z2)shr 15;
            w3:=z2-w4;
            pc.b:=
            (Col1.b*w1+PFColor(Integer(Col1)+3).b*w2+
            Col2.b*w3+PFColor(Integer(Col2)+3).b*w4)shr 15;
            pc.g:=
            (Col1.g*w1+PFColor(Integer(Col1)+3).g*w2+
            Col2.g*w3+PFColor(Integer(Col2)+3).g*w4)shr 15;
            pc.r:=
            (Col1.r*w1+PFColor(Integer(Col1)+3).r*w2+
            Col2.r*w3+PFColor(Integer(Col2)+3).r*w4)shr 15;
            Inc(pc);
            Inc(xP,xP2);
        end;
        Inc(yP,yP2);
    end;
end;

procedure TFastBMP.CopyRect(Dst:TFastRGB;x,y,w,h,sx,sy:Integer);
var
    n1,n2   : Pointer;
    i       : Integer;
begin
    if x<0 then
    begin
        Dec(sx,x);
        Inc(w,x);
        x:=0;
    end;

    if y<0 then
    begin
        Dec(sy,y);
        Inc(h,y);
        y:=0;
    end;

    if sx<0 then
    begin
        Dec(x,sx);
        Inc(w,sx);
        sx:=0;
    end;

    if sy<0 then
    begin
        Dec(y,sy);
        Inc(h,sy);
        sy:=0;
    end;

    if (sx>Width-1)or(sy>Height-1) then
        Exit;

    if (sx+w>Width) then
        Dec(w,(sx+w)-(Width));

    if (sy+h>Height) then
        Dec(h,(sy+h)-(Height));

    if (x+w>Dst.Width) then
        Dec(w,(x+w)-(Dst.Width));

    if (y+h>Dst.Height) then
        Dec(h,(y+h)-(Dst.Height));

    n1:=@Dst.Pixels[y,x];
    n2:=@Pixels[sy,sx];

    for i:=0 to h-1 do
    begin
        CopyMemory(n1,n2,w*3);
        n1:=Ptr(Integer(n1)+Dst.RowInc);
        n2:=Ptr(Integer(n2)+RowInc);
    end;
end;

procedure TFastBMP.Tile(Dst:TFastRGB);
var
    w,h,cy,cx   : Integer;
begin
    if ((Dst.Width=0)or(Dst.Height=0)or(Width=0)or(Height=0)) then
        Exit;

    CopyRect(Dst,0,0,Width,Height,0,0);
    w:=Width;
    h:=Height;
    cx:=Dst.Width;
    cy:=Dst.Height;

    while w<cx do
    begin
        Dst.CopyRect(Dst,w,0,w*2,h,0,0);
        Inc(w,w);
    end;

    while h<cy do
    begin
        Dst.CopyRect(Dst,0,h,w,h*2,0,0);
        Inc(h,h);
    end;
end;

// Vit's function, accurate to 256^3!
function TFastBMP.CountColors:Longint;
type
    TSpace = array[0..255,0..255,0..31]of Byte;
    PSpace = ^TSpace;
var
    x,y     : Integer;
    Tmp     : PFColor;
    Count   : Longint;
    Buff    : PByte;
    Test    : Byte;
    Space   : PSpace;
begin
    New(Space);
    FillChar(Space^,SizeOf(TSpace),0);
    Tmp:=Bits;
    Count:=0;

    for y:=0 to Height-1 do
    begin
        for x:=0 to Width-1 do
        begin
            Buff:=@Space^[Tmp.r,Tmp.g,Tmp.b shr 3];
            Test:=(1 shl(Tmp.b and 7));

            if(Buff^ and Test)=0 then
            begin
                Inc(Count);
                Buff^:=Buff^ or Test;
            end;

            Inc(Tmp);
        end;
        Tmp:=Ptr(Integer(Tmp)+Gap);
    end;

    Result:=Count;
    Dispose(Space);
end;


////////////////////////////////////////////////////////////////////////////////
// FAST256 _____________________________________________________________________
constructor TFast256.Create;
begin
    // only initializes some static properties:
    hDC:=CreateCompatibleDC(0);
    Width:=0; Height:=0; Pixels:=nil;
    bmInfo.bmiHeader.biSize:=SizeOf(TBitmapInfoHeader);
    bmInfo.bmiHeader.biPlanes:=1;
    bmInfo.bmiHeader.biBitCount:=8;
    bmInfo.bmiHeader.biCompression:=BI_RGB;
    Colors:=@bmInfo.bmiColors;
end;

destructor TFast256.Destroy;
begin
    DeleteDC(hDC);
    DeleteObject(Handle);
    ReallocMem(Pixels,0);
    inherited Destroy;
end;

procedure TFast256.SetSize(fWidth,fHeight:Integer);
var
    x   : Longint;
    i   : Integer;
begin
    if ((fWidth=Width)and(fHeight=Height)) then
        Exit;

    Width:=fWidth;
    Height:=fHeight;
    bmInfo.bmiHeader.biWidth:=Width;
    bmInfo.bmiHeader.biHeight:=Height;
    Handle:=CreateDIBSection(0,PBitmapInfo(@bmInfo)^,DIB_RGB_COLORS,Bits,hSection,0);
    DeleteObject(SelectObject(hDC,Handle));
    RowInc:=((Width*8+31)shr 5)shl 2;
    Gap:=RowInc-Width;
    Size:=RowInc*Height;
    ReallocMem(Pixels,Height*4);
    x:=Longint(Bits);

    for i:=0 to fHeight-1 do
    begin
        Pixels[i]:=Pointer(x);
        Inc(x,RowInc);
    end;
end;

procedure TFast256.LoadFromFile(FileName:string;cm:TConversionMode);
var
    hFile   : Integer;
    iRead   : Cardinal;
    tmp     : TFastBMP;
    fData   : TofStruct;
    fHead   : TBitmapFileHeader;
    fInfo   : TBitmapInfo256;
    fBits   : PBytes;
begin
    hFile:=OpenFile(PChar(FileName),fData,OF_READ);
    ReadFile(hFile,fHead,SizeOf(fHead),iRead,nil);
    ReadFile(hFile,fInfo,SizeOf(fInfo),iRead,nil);
    if fInfo.bmiheader.biBitCount<9 then
    begin
        SetSize(fInfo.bmiHeader.biWidth,Abs(fInfo.bmiHeader.biHeight));
        SetDIBColorTable(hDC,0,256,fInfo.bmiColors);
        bmInfo.bmiColors:=fInfo.bmiColors;
        SetFilePointer(hFile,fHead.bfOffBits,nil,FILE_BEGIN);
        GetMem(fBits,Size);
        ReadFile(hFile,fBits^,Size,iRead,nil);
        StretchDIBits(hDC,0,0,Width,Height,0,0,Width,Height,fBits,PBitmapInfo(@fInfo)^,DIB_RGB_COLORS,SRCCOPY);
        FreeMem(fBits);
        CloseHandle(hFile);
    end
    else
    begin
        CloseHandle(hFile);
        tmp:=TFastBMP.Create;
        tmp.LoadFromFile(FileName);
        LoadFromRGB(tmp,cm);
        tmp.free;
    end;
end;

procedure TFast256.SaveToFile(FileName:string);
var
    hFile   : Integer;
    iWrote  : Cardinal;
    fData   : TofStruct;
    fHead   : TBitmapFileHeader;
begin
    // saves upside down.. just Flop from FastFX if you want
    hFile:=OpenFile(PChar(FileName),fData,OF_CREATE or OF_WRITE);
    fHead.bfType:=19778; // "BM"
    fHead.bfSize:=Size+SizeOf(fHead)+SizeOf(bmInfo);
    fHead.bfOffBits:=SizeOf(fHead)+SizeOf(bmInfo);
    WriteFile(hFile,fHead,SizeOf(fHead),iWrote,nil);
    bmInfo.bmiHeader.biHeight:=Abs(bmInfo.bmiHeader.biHeight);
    WriteFile(hFile,bmInfo,SizeOf(bmInfo),iWrote,nil);
    bmInfo.bmiHeader.biHeight:=-bmInfo.bmiHeader.biHeight;
    WriteFile(hFile,PBytes(Bits)^,Size,iWrote,nil);
    CloseHandle(hFile);
end;

// if hBmp.bpp>8 then LoadFromRGB is used
procedure TFast256.LoadFromhBmp(hBmp:Integer;cm:TConversionMode);
var
    Bmp     : TBitmap;
    memDC   : Integer;
    Tmp     : TFastBMP;
begin
    GetObject(hBmp,SizeOf(Bmp),@Bmp);
    if Bmp.bmBitsPixel<9 then
    begin
        SetSize(Bmp.bmWidth,Bmp.bmHeight);
        memDC:=CreateCompatibleDC(0);
        SelectObject(memDC,hBmp);
        GetDIBits(memDC,hBmp,0,Height,Bits,PBitmapInfo(@bmInfo)^,DIB_RGB_COLORS);
        DeleteDC(memDC);
    end
    else
    begin
        Tmp:=TFastBMP.Create;
        Tmp.LoadFromhBmp(hBmp);
        LoadFromRGB(Tmp,cm);
        Tmp.Free;
    end;
end;

procedure TFast256.LoadFromRes(hInst:Integer;sName:string;cm:TConversionMode);
var
    hBmp    : Integer;
begin
    hBmp:=LoadImage(hInst,PChar(sName),IMAGE_BITMAP,0,0,LR_CREATEDIBSECTION);
    LoadFromhBmp(hBmp,cm);
    DeleteObject(hBmp);
end;

procedure TFast256.LoadFromRGBExact(FB:TFastRGB);
type
    TSpace = array[0..255,0..255,0..255]of Word;
    PSpace =^TSpace;
var
    x,y,i   : Integer;
    pc      : PFColor;
    pb      : PByte;
    a       : PWord;
    Space   : PSpace;
begin
    New(Space);
    FillChar(Space^,SizeOf(TSpace),255);
    SetSize(FB.Width,FB.Height);
    pb:=Bits; pc:=FB.Bits;
    i:=0;
    
    for y:=0 to FB.Height-1 do
    begin
        for x:=0 to FB.Width-1 do
        begin
            a:=@Space^[pc.b,pc.g,pc.r];
            if a^=$FFFF then
            begin
                a^:=i;
                Colors[i].b:=pc.b;
                Colors[i].g:=pc.g;
                Colors[i].r:=pc.r;
                Inc(i);
            end;
            pb^:=a^;
            Inc(pb);
            Inc(pc);
        end;
        pc:=Ptr(Integer(pc)+FB.Gap);
        pb:=Ptr(Integer(pb)+Gap);
    end;
    Dispose(Space);
end;

procedure TFast256.MakeGrayscale(FB:TFastRGB);
var
    x,y,i   : Integer;
    p       : PByte;
    Tmp     : PFColor;
    Div3    : array[0..765]of Byte;
begin
    SetSize(FB.Width,FB.Height);
    x:=0; y:=0;

    for i:=0 to 255 do
    begin
        Div3[x]:=y; Inc(x);
        Div3[x]:=y; Inc(x);
        Div3[x]:=y; Inc(x);
        Inc(y);
    end;

    Tmp:=FB.Bits;
    p:=Bits;

    for y:=0 to FB.Height-1 do
    begin
        for x:=0 to FB.Width-1 do
        begin
            p^:=Div3[Tmp.b+Tmp.g+Tmp.r];
            Inc(Tmp);
            Inc(p);
        end;
        
        Tmp:=Ptr(Integer(Tmp)+FB.Gap);
        p:=Ptr(Integer(p)+Gap);
    end;

    for i:=0 to 255 do
    begin
        bmInfo.bmiColors[i].r:=i;
        bmInfo.bmiColors[i].g:=i;
        bmInfo.bmiColors[i].b:=i;
    end;
end;

// median-cut color quantization implemented by Vit - THANK YOU! ;-)
// FS dither now available!
procedure TFast256.LoadFromRGB(FB:TFastRGB;cm:TConversionMode);
type
    TThreeInt   = record r,g,b:Integer;end;
    PThreeInt   = ^TThreeInt;
    TLineErrors = array [-1..-1] of TThreeInt;
    PLineErrors = ^TLineErrors;
    PCube       = ^TCube;
    TCube       = record
        X1,Y1,Z1,
        X2,Y2,Z2:   Byte;
    end;
var
    A,B,C,
    D,E,F   : Integer;
    Space   : array[0..31,0..31,0..31]of Byte;
    Cubes   : array[0..255]of TCube;
    Cl      : PFColor;
    IC      : PFColorEntry;
    CNum    : Byte;
    PB      : PByte;
    Cub     : PCube;
    Lines   : array[0..1]of PLineErrors;
    Ln,Nxt  : Integer;
    PTI     : PThreeInt;
    Dir     : ShortInt;

    procedure OptimizeCube(Cb:PCube);
    var
        F,G,H,I : ShortInt;
    begin
        A:=32;
        B:=32;
        C:=32;
        D:=-1;
        E:=-1;
        F:=-1;

        for G:=Cb.X1 to Cb.X2 do
            for H:=Cb.Y1 to Cb.Y2 do
                for I:=Cb.Z1 to Cb.Z2 do
                begin
                    if Space[G,H,I] > 0 then
                    begin
                        if G<A then A:=G;
                        if G>D then D:=G;
                        if H<B then B:=H;
                        if H>E then E:=H;
                        if I<C then C:=I;
                        if I>F then F:=I;
                    end;
                end;
                
        Cb.X1:=A;
        Cb.Y1:=B;
        Cb.Z1:=C;
        Cb.X2:=D;
        Cb.Y2:=E;
        Cb.Z2:=F;
    end;

begin

    if cm=cmGray then
    begin
        MakeGrayscale(FB);
        Exit;
    end;

    if gTool.Check256(FB) then
    begin
        LoadFromRGBExact(FB);
        Exit;
    end;

    SetSize(FB.Width,FB.Height);
    ZeroMemory(@Space,SizeOf(Space));

    Cl:=FB.Bits;
    for A:=0 to FB.Height-1 do
    begin
        for B:=0 to FB.Width-1 do
        begin
            Space[Cl.r shr 3,Cl.g shr 3,Cl.b shr 3]:=1;
            Inc(Cl);
        end;
        Cl:=Pointer(Integer(Cl)+FB.Gap);
    end;

    Cubes[0].X1:=0;
    Cubes[0].Y1:=0;
    Cubes[0].Z1:=0;
    Cubes[0].X2:=31;
    Cubes[0].Y2:=31;
    Cubes[0].Z2:=31;
    OptimizeCube(@Cubes[0]);
    CNum:=0;

    // search for biggest cube, divide that cube by largest side
    // until there are 255 cubes
    repeat
        A:=1;
        for B:=0 to CNum do
        begin
            Cub:=@Cubes[B];

            if Cub^.X2-Cub^.X1>A then
            begin
                A:=Cub^.X2-Cub^.X1;
                D:=0;
                F:=B;
            end;

            if Cub^.Y2-Cub^.Y1>A then
            begin
                A:=Cub^.Y2-Cub^.Y1;
                D:=1;
                F:=B;
            end;

            if Cub^.Z2-Cub^.Z1>A then
            begin
                A:=Cub^.Z2-Cub^.Z1;
                D:=2;
                F:=B;
            end;
        end;

        if A=1 then Break;

        Inc(CNum);
        Cubes[CNum]:=Cubes[F];
        Cub:=@Cubes[F];

        case D of
            0: begin
                Cub^.X2:=(Cub^.X1+Cub^.X2)shr 1;
                Cubes[CNum].X1:=Cub^.X2+1;
            end;

            1: begin
                Cub^.Y2:=(Cub^.Y1+Cub^.Y2)shr 1;
                Cubes[CNum].Y1:=Cub^.Y2+1;
            end;

            2: begin
                Cub^.Z2:=(Cub^.Z1+Cub^.Z2)shr 1;
                Cubes[CNum].Z1:=Cub^.Z2+1;
            end;
        end;

        OptimizeCube(@Cubes[CNum]);
        OptimizeCube(Cub);
    until CNum=255;

    if (cm=cmDither) then
    begin
        FillMemory(@Space,SizeOf(Space),255);

        for D:=0 to CNum do
        begin

            for A:=Cubes[D].X1 to Cubes[D].X2 do
                for B:=Cubes[D].Y1 to Cubes[D].Y2 do
                    for C:=Cubes[D].Z1 to Cubes[D].Z2 do
                    Space[A,B,C]:=D;

            Colors[D].r:=(Cubes[D].X1+Cubes[D].X2)shl 2;
            Colors[D].g:=(Cubes[D].Y1+Cubes[D].Y2)shl 2;
            Colors[D].b:=(Cubes[D].Z1+Cubes[D].Z2)shl 2;
        end;

        GetMem(Lines[0],24*(FB.Width+2));
        GetMem(Lines[1],24*(FB.Width+2));

        PB:=Bits;
        Cl:=FB.Bits;

        for B:=0 to FB.Width-1 do
        begin
            Lines[0,B].r:=Cl.r*16;
            Lines[0,B].g:=Cl.g*16;
            Lines[0,B].b:=Cl.b*16;
            Inc(Cl);
        end;

        Cl:=Pointer(Integer(Cl)+FB.Gap);
        Dir:=1;

        for A:=1 to FB.Height do
        begin
            Nxt:=A mod 2;
            Ln:=1-Nxt;
            if A<FB.Height then
            begin
                for B:=0 to FB.Width-1 do
                begin
                    Lines[Nxt,B].r:=Cl.r*16;
                    Lines[Nxt,B].g:=Cl.g*16;
                    Lines[Nxt,B].b:=Cl.b*16;
                    Inc(Cl);
                end;
                Cl:=Pointer(Integer(Cl)+FB.Gap);
            end;

            B:=0;

            if (Dir=-1) then
                B:=FB.Width-1;

            PTI:=@Lines[Ln,B];
            PB:=@Pixels[A-1,B];

            while((B>-1)and(B<FB.Width))do
            begin
                PTI.r:=PTI.r div 16;
                if      PTI.r>255 then PTI.r:=255
                else if PTI.r<0   then PTI.r:=0;
                PTI.g:=PTI.g div 16;
                if      PTI.g>255 then PTI.g:=255
                else if PTI.g<0   then PTI.g:=0;
                PTI.b:=PTI.b div 16;
                if      PTI.b>255 then PTI.b:=255
                else if PTI.b<0   then PTI.b:=0;

                F:=Space[PTI.r div 8,PTI.g div 8,PTI.b div 8];

                if F=255 then
                begin
                    D:=$0FFFFFFF;
                    IC:=@Colors[CNum];

                    for C:=CNum downto 0 do
                    begin
                        E:=Sqr(PTI.r-IC.r)+Sqr(PTI.g-IC.g)+Sqr(PTI.b-IC.b);

                        if E<D then
                        begin
                            D:=E;
                            F:=C;
                        end;

                        Dec(IC);
                    end;
                    Space[PTI.r div 8,PTI.g div 8,PTI.b div 8]:=F;
                end;

                PB^:=F;

                C:=PTI.r-Colors[F].r;

                if C<>0 then
                begin
                    Inc(Lines[Ln,B+Dir].r,C*7);
                    Inc(Lines[Nxt,B-Dir].r,C*3);
                    Inc(Lines[Nxt,B].r,C*5);
                    Inc(Lines[Nxt,B+Dir].r,C);
                end;

                C:=PTI.g-Colors[F].g;

                if C<>0 then
                begin
                    Inc(Lines[Ln,B+Dir].g,C*7);
                    Inc(Lines[Nxt,B-Dir].g,C*3);
                    Inc(Lines[Nxt,B].g,C*5);
                    Inc(Lines[Nxt,B+Dir].g,C);
                end;

                C:=PTI.b-Colors[F].b;

                if C<>0 then
                begin
                    Inc(Lines[Ln,B+Dir].b,C*7);
                    Inc(Lines[Nxt,B-Dir].b,C*3);
                    Inc(Lines[Nxt,B].b,C*5);
                    Inc(Lines[Nxt,B+Dir].b,C);
                end;

                Inc(PB,Dir);
                Inc(PTI,Dir);
                Inc(B,Dir);
            end;

            PB:=Pointer(Integer(PB)+Gap);
            Dir:=-Dir;
        end;
        Dispose(Lines[0]);
        Dispose(Lines[1]);
    end
    else // no dither
    begin

        for D:=0 to CNum do
        begin
            for A:=Cubes[D].X1 to Cubes[D].X2 do
                for B:=Cubes[D].Y1 to Cubes[D].Y2 do
                    for C:=Cubes[D].Z1 to Cubes[D].Z2 do
                        Space[A,B,C]:=D;

            Colors[D].r:=(Cubes[D].X1+Cubes[D].X2)shl 2;
            Colors[D].g:=(Cubes[D].Y1+Cubes[D].Y2)shl 2;
            Colors[D].b:=(Cubes[D].Z1+Cubes[D].Z2)shl 2;
        end;

        PB:=Bits;
        Cl:=FB.Bits;

        for A:=0 to FB.Height-1 do
        begin
            for B:=0 to FB.Width-1 do
            begin
                PB^:=Space[Cl^.r shr 3,Cl^.g shr 3,Cl^.b shr 3];
                Inc(PB);
                Inc(Cl);
            end;
            PB:=Pointer(Integer(PB)+Gap);
            Cl:=Pointer(Integer(Cl)+FB.Gap);
        end;
    end;
end;

procedure TFast256.SetInterface(fWidth,fHeight:Integer;pBits:Pointer);
var
    x   : Longint;
    i   : Integer;
begin
    Width:=fWidth;
    Height:=Abs(fHeight);
    bmInfo.bmiHeader.biWidth:=fWidth;
    bmInfo.bmiHeader.biHeight:=fHeight;
    RowInc:=((Width*8+31)shr 5)shl 2;
    Gap:=RowInc-Width;
    Size:=RowInc*Height;
    ReallocMem(Pixels,Height*4);
    Bits:=pBits;
    x:=Longint(Bits);

    for i:=0 to fHeight-1 do
    begin
        Pixels[i]:=Pointer(x);
        Inc(x,RowInc);
    end;
end;

procedure TFast256.Copy(FB:TFast256);
begin
    SetSize(FB.Width,FB.Height);
    CopyMemory(Bits,FB.Bits,Size);
    bmInfo.bmiColors:=FB.bmInfo.bmiColors;
end;

procedure TFast256.ShiftColors(Amount:Integer);
var
    Buf : Pointer;
begin
  if (Amount<0) then
    Amount:=256-(Abs(Amount) mod 256);

  if (Amount>256) then
    Amount:=Amount mod 256;

  if (Amount=0) then
    Exit;

  GetMem(Buf,Amount*4);
  CopyMemory(Buf,Ptr(Integer(@bmInfo.bmiColors)+((256-Amount)*4)),Amount*4);
  MoveMemory(Ptr(Integer(@bmInfo.bmiColors)+(Amount*4)),@bmInfo.bmiColors,(256-Amount)*4);
  CopyMemory(@bmInfo.bmiColors,Buf,Amount*4);
  FreeMem(Buf);
end;

procedure TFast256.RGB(ra,ga,ba:Integer);
var
    Tmp : PFColorEntry;
    i   : Byte;
begin
    Tmp:=@bmInfo.bmiColors[0];

    for i:=0 to 255 do
    begin
        Tmp.r := gTool.IntToByte(Tmp.r+ra);
        Tmp.g := gTool.IntToByte(Tmp.g+ga);
        Tmp.b := gTool.IntToByte(Tmp.b+ba);
        Inc(Tmp);
    end;
end;

procedure TFast256.Contrast(Amount:Integer);
var
    Table   : array[0..255]of Byte;
    Tmp     : PFColorEntry;
    x       : Integer;
    i       : Byte;
begin
    for i:=0 to 126 do
    begin
        x:=(Abs(128-i)*Amount)div 256;
        Table[i] := gTool.IntToByte(i-x);
    end;

    for i:=127 to 255 do
    begin
        x:=(Abs(128-i)*Amount)div 256;
        Table[i] := gTool.IntToByte(i+x);
    end;

    Tmp:=@bmInfo.bmiColors[0];

    for i:=0 to 255 do
    begin
        Tmp.r:=Table[Tmp.r];
        Tmp.g:=Table[Tmp.g];
        Tmp.b:=Table[Tmp.b];
        Inc(Tmp);
    end;
end;

procedure TFast256.Saturation(Amount:Integer);
var
    Grays   : array[0..767]of Integer;
    Alpha   : array[0..255]of Word;
    Tmp     : PFColorEntry;
    a,b     : Integer;
    i       : Byte;
begin
    for i:=0 to 255 do
        Alpha[i]:=(i*Amount)shr 8;

    a:=0;

    for i:=0 to 255 do
    begin
        b:=i-Alpha[i];
        Grays[a]:=b; Inc(a);
        Grays[a]:=b; Inc(a);
        Grays[a]:=b; Inc(a);
    end;

    Tmp:=@bmInfo.bmiColors[0];

    for i:=0 to 255 do
    begin
        a:=Grays[Tmp.r+Tmp.g+Tmp.b];
        Tmp.r := gTool.IntToByte(a+Alpha[Tmp.r]);
        Tmp.g := gTool.IntToByte(a+Alpha[Tmp.g]);
        Tmp.b := gTool.IntToByte(a+Alpha[Tmp.b]);
        Inc(Tmp);
    end;
end;

procedure TFast256.Lightness(Amount:Integer);
var
    Table   : array[0..255]of Byte;
    Tmp     : PFColorEntry;
    i       : Byte;
begin
    if Amount<0 then
    begin
        Amount:=-Amount;

        for i:=0 to 255 do
            Table[i] := gTool.IntToByte(i-((Amount*i)shr 8));
    end else
        for i:=0 to 255 do
            Table[i] := gTool.IntToByte(i+((Amount*(i xor 255))shr 8));

    Tmp:=@bmInfo.bmiColors[0];

    for i:=0 to 255 do
    begin
        Tmp.r:=Table[Tmp.r];
        Tmp.g:=Table[Tmp.g];
        Tmp.b:=Table[Tmp.b];
        Inc(Tmp);
    end;
end;

procedure TFast256.FillColors(c1,c255:TFColor);
var
    ir,
    ig,
    ib,
    r,
    g,
    b       : Single;
    Tmp     : PFColorEntry;
    i       : Byte;
begin
    ir:=(c255.r-c1.r)/256;
    r:=c1.r;

    ig:=(c255.g-c1.g)/256;
    g:=c1.g;

    ib:=(c255.b-c1.b)/256;
    b:=c1.b;
    
    Tmp:=@bmInfo.bmiColors[1];

    for i:=0 to 253 do
    begin
        r:=r+ir; Tmp.r:=Round(r);
        g:=g+ig; Tmp.g:=Round(g);
        b:=b+ib; Tmp.b:=Round(b);
        Inc(Tmp);
    end;

    bmInfo.bmiColors[0]   := gTool.ColorToEntry(c1);
    bmInfo.bmiColors[255] := gTool.ColorToEntry(c255);
end;

function TFast256.MakePalette: HPALETTE;
type
    TPalette256=record
        Ver,
        Entries: Word;
        Colors:  array[0..255]of TPaletteEntry;
    end;
var
    Pal : TPalette256;
    i   : Byte;
begin
    Pal.Ver:=$300;
    Pal.Entries:=256;
    
    for i:=0 to 255 do
    begin
        Pal.Colors[i].peRed:=Colors[i].r;
        Pal.Colors[i].peGreen:=Colors[i].g;
        Pal.Colors[i].peBlue:=Colors[i].b;
    end;

    Result:=CreatePalette(PLogPalette(@pal)^);
end;

procedure TFast256.Draw(fdc,x,y:Integer);
begin
    StretchDIBits(fdc,x,y,Width,Height,0,0,Width,Height,Bits,PBitmapInfo(@bmInfo)^,DIB_RGB_COLORS,SRCCOPY);
end;

procedure TFast256.Stretch(fdc,x,y,w,h:Integer);
begin
    SetStretchBltMode(fdc,STRETCH_DELETESCANS);
    StretchDIBits(fdc,x,y,w,h,0,0,Width,Height,Bits,PBitmapInfo(@bmInfo)^,DIB_RGB_COLORS,SRCCOPY);
end;

procedure TFast256.DrawRect(fdc,x,y,w,h,sx,sy:Integer);
begin
    StretchDIBits(fdc,x,y,w,h,sx,sy,w,h,Bits,PBitmapInfo(@bmInfo)^,DIB_RGB_COLORS,SRCCOPY);
end;

procedure TFast256.StretchRect(fdc,x,y,w,h,sx,sy,sw,sh:Integer);
begin
    SetStretchBltMode(fdc,STRETCH_DELETESCANS);
    StretchDIBits(fdc,x,y,w,h,sx,sy,sw,sh,Bits,PBitmapInfo(@bmInfo)^,DIB_RGB_COLORS,SRCCOPY);
end;

procedure TFast256.TileDraw(fdc,x,y,w,h:Integer);
var
    wd,hd,
    hBmp,
    memDC: Integer;
begin
    if(Width=0)or(Height=0)then
        Exit;
        
    memDC:=CreateCompatibleDC(fdc);
    hBmp:=CreateCompatibleBitmap(fdc,w,h);
    SelectObject(memDC,hBmp);
    Draw(memDC,0,0);
    wd:=Width;
    hd:=Height;

    while wd<w do
    begin
        BitBlt(memDC,wd,0,wd*2,h,memDC,0,0,SRCCOPY);
        Inc(wd,wd);
    end;

    while hd<h do
    begin
        BitBlt(memDC,0,hd,w,hd*2,memDC,0,0,SRCCOPY);
        Inc(hd,hd);
    end;
    
    BitBlt(fdc,x,y,w,h,memDC,0,0,SRCCOPY);
    DeleteDC(memDC);
    DeleteObject(hBmp);
end;

procedure TFast256.CopyRect(Dst:TFastRGB;x,y,w,h,sx,sy:Integer);
var
    n1,
    n2  : Pointer;
    i   :     Integer;
begin
    if x<0 then
    begin
        Dec(sx,x);
        Inc(w,x);
        x:=0;
    end;

    if y<0 then
    begin
        Dec(sy,y);
        Inc(h,y);
        y:=0;
    end;

    if sx<0 then
    begin
        Dec(x,sx);
        Inc(w,sx);
        sx:=0;
    end;

    if sy<0 then
    begin
        Dec(y,sy);
        Inc(h,sy);
        sy:=0;
    end;

    if(sx>Width-1)or(sy>Height-1)then
        Exit;

    if (sx+w>Width) then
        Dec(w,(sx+w)-(Width));

    if (sy+h>Height) then
        Dec(h,(sy+h)-(Height));

    if (x+w>Dst.Width) then
        Dec(w,(x+w)-(Dst.Width));

    if (y+h>Dst.Height) then
        Dec(h,(y+h)-(Dst.Height));

    n1:=@TFast256(Dst).Pixels[y,x];
    n2:=@Pixels[sy,sx];
    for i:=0 to h-1 do
    begin
        CopyMemory(n1,n2,w);
        n1:=Ptr(Integer(n1)+Dst.RowInc);
        n2:=Ptr(Integer(n2)+RowInc);
    end;
end;

procedure TFast256.Tile(Dst:TFastRGB);
var
    w,
    h,
    cy,
    cx  : Integer;
begin
    if(Dst.Width=0)or(Dst.Height=0)or(Width=0)or(Height=0)then
        Exit;
        
    CopyRect(Dst,0,0,Width,Height,0,0);
    w:=Width;
    h:=Height;
    cx:=Dst.Width;
    cy:=Dst.Height;

    while w<cx do
    begin
        Dst.CopyRect(Dst,w,0,w*2,h,0,0);
        Inc(w,w);
    end;

    while h<cy do
    begin
        Dst.CopyRect(Dst,0,h,w,h*2,0,0);
        Inc(h,h);
    end;
end;

procedure TFast256.Resize(Dst:TFastRGB);
var
    xCount,
    yCount,
    x,y,xP,yP,
    xD,yD,
    yiScale,
    xiScale     : Integer;
    xScale,
    yScale      : Single;
    Read,
    Line        : PBytes;
    Tmp         : Byte;
    pc          : PByte;
begin
    if(Dst.Width=0)or(Dst.Height=0)then
        Exit;

    if(Dst.Width=Width)and(Dst.Height=Height)then
    begin
        CopyMemory(Dst.Bits,Bits,Size);
        Exit;
    end;

    xScale:=Dst.Width/Width;
    yScale:=Dst.Height/Height;

    if(xScale<1)or(yScale<1)then
    begin  // shrinking
        xiScale:=(Width shl 16) div Dst.Width;
        yiScale:=(Height shl 16) div Dst.Height;
        yP:=0;
        for y:=0 to Dst.Height-1 do
        begin
            xP:=0;
            read:=Pixels[yP shr 16];
            pc:=@TFast256(Dst).Pixels[y,0];

            for x:=0 to Dst.Width-1 do
            begin
                pc^:=Read[xP shr 16];
                Inc(pc);
                Inc(xP,xiScale);
            end;

            Inc(yP,yiScale);
        end;
    end
    else   // zooming
    begin
        yiScale:=Round(yScale+0.5);
        xiScale:=Round(xScale+0.5);
        GetMem(Line,Dst.Width);
        for y:=0 to Height-1 do
        begin
            yP:=Trunc(yScale*y);
            Read:=Pixels[y];
            for x:=0 to Width-1 do
            begin
                xP:=Trunc(xScale*x);
                Tmp:=Read[x];
                for xCount:=0 to xiScale-1 do
                begin
                    xD:=xCount+xP;
                    if xD>=Dst.Width then Break;
                    Line[xD]:=Tmp;
                end;
            end;

            for yCount:=0 to yiScale-1 do
            begin
                yD:=yCount+yP;
                if yD>=Dst.Height then Break;
                CopyMemory(TFast256(Dst).Pixels[yD],Line,Dst.Width);
            end;
        end;
        FreeMem(Line);
    end;
end;

// this will only look good if the color table is evenly spaced
procedure TFast256.SmoothResize(Dst:TFastRGB);
var
    x,y,xP,yP,
    yP2,xP2,
    t,z,z2,iz2,
    w1,w2,w3,w4     : Integer;
    Read,Read2      : PBytes;
    pc,Col1,Col2    : PByte;
begin
    if(Dst.Width<1)or(Dst.Height<1)then
        Exit;

    if Width=1 then
    begin
        Resize(Dst);
        Exit;
    end;

    if(Dst.Width=Width)and(Dst.Height=Height)then
    begin
        CopyMemory(Dst.Bits,Bits,Size);
        Exit;
    end;

    xP2:=((Width-1)shl 15)div Dst.Width;
    yP2:=((Height-1)shl 15)div Dst.Height;
    yP:=0;

    for y:=0 to Dst.Height-1 do
    begin
        xP:=0;
        Read:=Pixels[yP shr 15];
        if yP shr 16<Height-1 then
            Read2:=Pixels[yP shr 15+1]
        else
            Read2:=Pixels[yP shr 15];

        pc:=@TFast256(Dst).Pixels[y,0];
        z2:=yP and $7FFF;
        iz2:=$8000-z2;

        for x:=0 to Dst.Width-1 do
        begin
            t:=xP shr 15;
            Col1:=@Read[t];
            Col2:=@Read2[t];
            z:=xP and $7FFF;
            w2:=(z*iz2)shr 15;
            w1:=iz2-w2;
            w4:=(z*z2)shr 15;
            w3:=z2-w4;
            pc^:= (Col1^*w1+PByte(Integer(Col1)+1)^*w2+
                   Col2^*w3+PByte(Integer(Col2)+1)^*w4)shr 15;
            Inc(pc);
            Inc(xP,xP2);
        end;
        Inc(yP,yP2);
    end;
end;

// FastIMG _____________________________________________________________________

constructor TFastIMG.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);

    SetBounds(Left,Top,100,100);
end;

destructor TFastIMG.Destroy;
begin
    fBmp:=nil;

    inherited Destroy;
end;

procedure TFastIMG.Erase(var Msg:TWMEraseBkgnd);
begin
    // May I need something to do ????
end;

procedure TFastIMG.SetBitmap(x:TFastRGB);
begin
    fBmp:=x;

    if fBmp is TFast256 then
        fDataType:=dtFast256
    else
    if fBmp is TFastBMP then
        fDataType:=dtFastBMP;

    if fAutoSize then
        SetBounds(Left,Top,fBmp.Width,fBmp.Height);
end;

procedure TFastIMG.LoadFromFile(x:string);
begin
    if FileExists(x) then
    begin
        if not Assigned(fBmp) then
        begin
            case fDataType of
                dtFastBMP: fBmp:=TFastBMP.Create;
                dtFast256: fBmp:=TFast256.Create;
            end;
        end;

        case fDataType of
            dtFastBMP: TFastBMP(fBmp).LoadFromFile(x);
            dtFast256: TFast256(fBmp).LoadFromFile(x,fCMode);
        end;

        if fAutoSize then
            SetBounds(Left,Top,fBmp.Width,fBmp.Height);
    end;
end;

procedure TFastIMG.LoadFromResource(hInst:Integer;sName:string);
begin
    if not Assigned(fBmp) then
    begin
        case fDataType of
            dtFastBMP: fBmp:=TFastBMP.Create;
            dtFast256: fBmp:=TFast256.Create;
        end;
    end;

    case fDataType of
        dtFastBMP: TFastBMP(fBmp).LoadFromRes(hInst,sName);
        dtFast256: TFast256(fBmp).LoadFromRes(hInst,sName,cmDither);
    end;

    if fAutoSize then
        SetBounds(Left,Top,fBmp.Width,fBmp.Height);
end;

procedure TFastIMG.LoadFromStream(x:TStream);
begin
    // not availed.
end;

procedure TFastIMG.SetAutoSize(x:Boolean);
begin
    fAutoSize:=x;

    if fAutoSize and Assigned(Bmp)then
    begin
        Align:=alNone;
        SetBounds(Left,Top,Bmp.Width,Bmp.Height);
    end;
end;

procedure TFastIMG.SetSmthSize(x:Boolean);
begin
    fSmthSize:=x;
    Refresh;
end;

procedure TFastIMG.SetDrawStyle(x:TDrawStyle);
begin
    fDrawStyle:=x;
    Refresh;
end;

procedure TFastIMG.SetDataType(x:TDataType);
begin
    fDataType:=x;

    if csDesigning in ComponentState then
    begin
        if Assigned(fBmp) then
            fBmp.Free;

        case fDataType of
            dtFastBMP: fBmp:=TFastBMP.Create;
            dtFast256: fBmp:=TFast256.Create;
        end;

        if fFileName<>'' then LoadFromFile(fFileName);

        Refresh;
    end;
end;

procedure TFastIMG.SetConversionMode(x:TConversionMode);
begin
    fCMode:=x;

    if(csDesigning in ComponentState)and(fFileName<>'')then
        LoadFromFile(fFileName);

    Refresh;
end;

procedure TFastIMG.SetFileName(const x:TBMPFilename);
begin
    if x='' then
        Bmp:=nil;

    fFileName:=x;
    LoadFromFile(x);

    Refresh;
end;

procedure TFastIMG.Paint(var Msg:TWMPaint);
var
    ps      : TPaintStruct;
    r1,r2   : HRGN;
    ax,ay   : Single;
    iw,ih,
    cx,cy   : Integer;
    Tmp     : TFastRGB;
begin
    BeginPaint(Handle,ps);
    if Assigned(Bmp)then
        case fDrawStyle of
            dsDraw:
            begin
                if(Width>fBmp.Width)or(Height>fBmp.Height)then
                begin
                    r1:=CreateRectRgn(0,0,Width,Height);
                    r2:=CreateRectRgn(0,0,fBmp.Width,fBmp.Height);
                    CombineRgn(r1,r1,r2,RGN_XOR);
                    DeleteObject(r2);
                    FillRgn(ps.hdc,r1,Brush.Handle);
                    DeleteObject(r1);
                end;

                fBmp.Draw(ps.hdc,0,0);
            end;

            dsStretch:
            begin
                if fSmthSize then
                begin
                    case fDataType of
                        dtFastBMP:
                            Tmp:=TFastBMP.Create;

                        dtFast256:
                        begin
                            Tmp:=TFast256.Create;
                            TFast256(Tmp).Colors^:=TFast256(fBmp).Colors^;
                        end;
                    end;
                    
                    Tmp.SetSize(Width,Height);
                    fBmp.SmoothResize(Tmp);
                    Tmp.Draw(ps.hdc,0,0);
                    Tmp.Free;
                end else
                    fBmp.Stretch(ps.hdc,0,0,Width,Height);
            end;

            dsResize:
            begin
                if fBmp.Width=0 then
                    iw:=1
                else
                    iw:=fBmp.Width;

                if fBmp.Height=0 then
                    ih:=1
                else
                    ih:=fBmp.Height;

                ax:=Width/iw;
                ay:=Height/ih;

                if (ax>ay) then
                begin
                    ax:=fBmp.Width*ay;       ay:=Height;
                    cx:=Round((Width-ax)/2); cy:=0;
                end
                else
                begin
                    ay:=fBmp.Height*ax;       ax:=Width;
                    cy:=Round((Height-ay)/2); cx:=0;
                end;

                iw:=Round(ax); ih:=Round(ay);

                if(cx<>0)or(cy<>0)then
                begin
                    r1:=CreateRectRgn(0,0,Width,Height);
                    r2:=CreateRectRgn(cx,cy,iw+cx,ih+cy);
                    CombineRgn(r1,r1,r2,RGN_XOR);
                    DeleteObject(r2);
                    FillRgn(ps.hdc,r1,Brush.Handle);
                    DeleteObject(r1);
                end;

                if (fSmthSize) then
                begin
                    case fDataType of
                        dtFastBMP:
                        begin
                            Tmp:=TFastBMP.Create;
                            TFastBMP(Tmp).SetSize(iw,ih);
                        end;

                        dtFast256:
                        begin
                            Tmp:=TFast256.Create;
                            TFast256(Tmp).SetSize(iw,ih);
                            TFast256(Tmp).Colors^:=TFast256(fBmp).Colors^;
                        end;
                end;

                fBmp.SmoothResize(Tmp);
                Tmp.Draw(ps.hdc,cx,cy);
                Tmp.Free;
            end else
                fBmp.Stretch(ps.hdc,cx,cy,iw,ih);
            end;
            
            dsTile:
                fBmp.TileDraw(ps.hdc,0,0,Width,Height);

            dsCenter:
            begin
                cx:=(Width-fBmp.Width)div 2;
                cy:=(Height-fBmp.Height)div 2;

                if(Width>fBmp.Width)or(Height>fBmp.Height)then
                begin
                    r1:=CreateRectRgn(0,0,Width,Height);
                    r2:=CreateRectRgn(cx,cy,fBmp.Width+cx,fBmp.Height+cy);
                    CombineRgn(r1,r1,r2,RGN_XOR);
                    DeleteObject(r2);
                    FillRgn(ps.hdc,r1,Brush.Handle);
                    DeleteObject(r1);
                end;

                fBmp.Draw(ps.hdc,cx,cy);
        end;
    end
    else //fBmp=nil
        FillRect(ps.hdc,ps.rcPaint,Brush.Handle);

    if(csDesigning in ComponentState)then
        DrawFocusRect(ps.hdc,RECT(0,0,Width,Height));

    EndPaint(Handle,ps);
end;

begin
    gTool := TFastGraphic.Create;
end.
